import base64
import os
from datetime import datetime
from enum import Enum
from typing import Dict, Optional, Union

import aiofiles
import backoff
import httpx
import requests
from nuclia_models.common.utils import Aggregation
from nuclia_models.config.proto import ExtractConfig, SplitConfiguration
from nuclia_models.events.activity_logs import (  # type: ignore
    ActivityLogsAskQuery,
    ActivityLogsQuery,
    ActivityLogsSearchQuery,
    DownloadActivityLogsAskQuery,
    DownloadActivityLogsQuery,
    DownloadActivityLogsSearchQuery,
    DownloadFormat,
    EventType,
)
from nuclia_models.events.remi import RemiQuery
from nuclia_models.worker.tasks import TaskStartKB
from nucliadb_models.search import AskRequest, SummarizeRequest
from nucliadb_sdk import NucliaDB, NucliaDBAsync
from tqdm import tqdm

from nuclia.exceptions import RateLimitError
from nuclia.lib.utils import (
    USER_AGENT,
    build_httpx_async_client,
    build_httpx_client,
    handle_http_async_errors,
    handle_http_sync_errors,
)

RESOURCE_PATH = "/resource/{rid}"
RESOURCE_PATH_BY_SLUG = "/slug/{slug}"
SEARCH_PATH = "/search"
CREATE_RESOURCE_PATH = "/resources"
CREATE_VECTORSET = "/vectorset/{vectorset}"
VECTORSETS = "/vectorsets"
COUNTER = "/counters"
SEARCH_URL = "/search"
FIND_URL = "/find"
ASK_URL = "/ask"
SUMMARIZE_URL = "/summarize"
LABELS_URL = "/labelsets"
ENTITIES_URL = "/entitiesgroups"
DOWNLOAD_EXPORT_URL = "/export/{export_id}"
DOWNLOAD_URL = "/{uri}"
TUS_UPLOAD_RESOURCE_URL = "/resource/{rid}/file/{field}/tusupload"
TUS_UPLOAD_URL = "/tusupload"
ACTIVITY_LOG_URL = "/activity/{type}/query/download"
ACTIVITY_LOG_DOWNLOAD_REQUEST_URL = "/activity/download_request/{request_id}"
ACTIVITY_LOG_QUERY_URL = "/activity/{type}/query"
NOTIFICATIONS = "/notifications"
REMI_QUERY_URL = "/remi/query"
REMI_EVENT_URL = "/remi/events/{event_id}"
REMI_SCORES_URL = "/remi/scores"
LIST_TASKS = "/tasks"
START_TASK = "/task/start"
STOP_TASK = "/task/{task_id}/stop"
DELETE_TASK = "/task/{task_id}"
GET_TASK = "/task/{task_id}/inspect"
RESTART_TASK = "/task/{task_id}/restart"
EXTRACT_STRATEGIES = "/extract_strategies"
DELETE_EXTRACT_STRATEGY = "/extract_strategies/strategy/{id}"
SPLIT_STRATEGIES = "/split_strategies"
DELETE_SPLIT_STRATEGY = "/split_strategies/strategy/{id}"

DOWNLOAD_FORMAT_HEADERS = {
    DownloadFormat.CSV: "text/csv",
    DownloadFormat.NDJSON: "application/x-ndjson",
}


class Environment(str, Enum):
    CLOUD = "CLOUD"
    OSS = "OSS"


class BaseNucliaDBClient:
    environment: Environment
    base_url: str
    api_key: Optional[str]
    url: Optional[str] = None
    region: str
    headers: Dict[str, str]

    reader_headers: Dict[str, str]
    writer_headers: Dict[str, str]

    def __init__(
        self,
        *,
        environment: Environment = Environment.CLOUD,
        url: Optional[str] = None,
        api_key: Optional[str] = None,
        user_token: Optional[str] = None,
        region: Optional[str] = None,
        base_url: Optional[str] = None,
    ):
        if environment == Environment.OSS:
            self.region = "on-prem"
        else:
            if region is None:
                raise ValueError("region is required for cloud environment")
            self.region = region

        self.headers = {"User-Agent": USER_AGENT}
        if user_token is not None:
            self.headers["Authorization"] = f"Bearer {user_token}"
        self.headers["X-SYNCHRONOUS"] = "True"
        if base_url is not None:
            v2url = base_url
        if base_url is None and url is not None:
            v2url = "/".join(url.split("/")[:-3])
        self.base_url = v2url
        self.api_key = api_key
        self.environment = environment

        if url is not None:
            self.kbid = url.strip("/").split("/")[-1]
            self.url = url

        if environment == Environment.CLOUD and api_key is not None:
            self.reader_headers = {
                "X-NUCLIA-SERVICEACCOUNT": f"Bearer {api_key}",
            }
            self.writer_headers = {
                "X-NUCLIA-SERVICEACCOUNT": f"Bearer {api_key}",
                "X-SYNCHRONOUS": "True",
            }
        elif (
            environment == Environment.CLOUD
            and api_key is None
            and user_token is not None
        ):
            self.reader_headers = {
                "Authorization": f"Bearer {user_token}",
            }
            self.writer_headers = {
                "Authorization": f"Bearer {user_token}",
                "X-SYNCHRONOUS": "True",
            }
        elif (
            environment == Environment.CLOUD and api_key is None and user_token is None
        ):
            # Public
            self.reader_headers = {}
            self.writer_headers = {}
        else:
            self.reader_headers = {
                "X-NUCLIADB-ROLES": "READER",
            }
            self.writer_headers = {
                "X-NUCLIADB-ROLES": "WRITER",
                "X-SYNCHRONOUS": "True",
            }


class NucliaDBClient(BaseNucliaDBClient):
    reader_session: Optional[httpx.Client] = None
    stream_session: Optional[requests.Session] = None
    writer_session: Optional[httpx.Client] = None
    ndb: NucliaDB

    def __init__(
        self,
        *,
        environment: Environment = Environment.CLOUD,
        url: Optional[str] = None,
        api_key: Optional[str] = None,
        user_token: Optional[str] = None,
        region: Optional[str] = None,
        base_url: Optional[str] = None,
    ):
        super().__init__(
            environment=environment,
            url=url,
            api_key=api_key,
            user_token=user_token,
            region=region,
            base_url=base_url,
        )
        self.ndb = NucliaDB(
            region=self.region, url=self.base_url, api_key=api_key, headers=self.headers
        )

        if url is not None:
            self.reader_session = build_httpx_client(
                headers=self.reader_headers,
                base_url=url,  # type: ignore
            )
            self.stream_session = requests.Session()
            self.stream_session.headers.update(
                {"User-Agent": USER_AGENT, **self.reader_headers}
            )
            self.writer_session = build_httpx_client(
                headers=self.writer_headers,
                base_url=url,  # type: ignore
            )

    def __repr__(self):
        return f"{self.environment} - {self.url}"

    def notifications(self):
        if self.url is None or self.stream_session is None:
            raise Exception("KB not configured")
        url = f"{self.url}{NOTIFICATIONS}"
        response = self.stream_session.get(url, stream=True, timeout=3660)
        handle_http_sync_errors(response)
        return response

    def ask(self, request: AskRequest, timeout: int = 1000):
        if self.url is None or self.stream_session is None:
            raise Exception("KB not configured")
        url = f"{self.url}{ASK_URL}"
        response: requests.Response = self.stream_session.post(
            url,
            data=request.model_dump_json(),
            stream=True,
            timeout=timeout,
        )
        handle_http_sync_errors(response)
        return response

    def download_export(self, export_id: str, path: str, chunk_size: int):
        if self.reader_session is None:
            raise Exception("KB not configured")

        parent = os.path.dirname(path)
        if not os.path.exists(parent):
            os.makedirs(parent)

        status = self.ndb.export_status(kbid=self.kbid, export_id=export_id)
        export_size = status.total

        url = DOWNLOAD_EXPORT_URL.format(export_id=export_id)
        with self.reader_session.stream("GET", url) as response:
            with (
                open(path, "wb") as file,
                tqdm(
                    desc="Downloading data",
                    total=export_size,
                    unit="iB",
                    unit_scale=True,
                ) as pbar,
            ):
                for chunk in response.iter_bytes(chunk_size):
                    pbar.update(len(chunk))
                    file.write(chunk)

    def download(self, uri: str) -> bytes:
        # uri has format
        # /kb/2a00d5b4-cfcc-48eb-85ac-d70bfd38b26d/resource/41d02aac4ade48098b23e38141807738/file/file/download/field
        # we need to remove the kb url
        if self.reader_session is None:
            raise Exception("KB not configured")

        uri_parts = uri.split("/")
        if len(uri_parts) < 9:
            raise AttributeError("Not a valid download uri")

        new_uri = "/".join(uri_parts[3:])
        url = DOWNLOAD_URL.format(uri=new_uri)
        response: httpx.Response = self.reader_session.get(url)
        handle_http_sync_errors(response)
        return response.content

    @backoff.on_exception(
        backoff.expo,
        RateLimitError,
        jitter=backoff.random_jitter,
        max_tries=5,
        factor=10,
    )
    def start_tus_upload(
        self,
        size: int,
        filename: str,
        field: Optional[str] = None,
        rid: Optional[str] = None,
        md5: Optional[str] = None,
        content_type: str = "application/octet-stream",
        extract_strategy: Optional[str] = None,
        split_strategy: Optional[str] = None,
    ):
        if self.writer_session is None:
            raise Exception("KB not configured")

        if rid is not None:
            if field is None:
                field = filename
            url = TUS_UPLOAD_RESOURCE_URL.format(rid=rid, field=field)
        else:
            url = TUS_UPLOAD_URL
        encoded_filename = base64.b64encode(filename.encode()).decode()
        headers = {
            "upload-length": str(size),
            "tus-resumable": "1.0.0",
            "upload-metadata": f"filename {encoded_filename}",
            "content-type": content_type,
        }
        if md5 is not None:
            headers["upload-metadata"] += (
                f",md5 {base64.b64encode(md5.encode()).decode()}"
            )
        if extract_strategy is not None:
            headers["x-extract-strategy"] = extract_strategy

        if split_strategy is not None:
            headers["x-split-strategy"] = split_strategy

        response: httpx.Response = self.writer_session.post(url, headers=headers)
        handle_http_sync_errors(response)
        return response.headers.get("Location")

    def patch_tus_upload(self, upload_url: str, data: bytes, offset: int) -> int:
        if self.writer_session is None:
            raise Exception("KB not configured")

        headers = {
            "upload-offset": str(offset),
        }
        url = httpx.URL(upload_url)
        if url.is_relative_url:
            # Relative path starting with /kb/..., we remove /kb/kbid (the base_url of the session includes it)
            url = httpx.URL("/".join(url.path.split("/")[3:]))

        response: httpx.Response = self.writer_session.patch(
            url, headers=headers, content=data
        )
        handle_http_sync_errors(response)
        return int(response.headers.get("Upload-Offset"))

    def summarize(
        self,
        request: SummarizeRequest,
        extra_headers: Optional[dict[str, str]] = None,
        timeout: int = 1000,
    ):
        if self.url is None or self.writer_session is None:
            raise Exception("KB not configured")
        url = f"{self.url}{SUMMARIZE_URL}"
        assert self.reader_session
        response = self.reader_session.post(
            url, json=request.model_dump(), headers=extra_headers, timeout=timeout
        )
        handle_http_sync_errors(response)
        return response

    def logs_query(
        self,
        type: EventType,
        query: Union[ActivityLogsQuery, ActivityLogsSearchQuery, ActivityLogsAskQuery],
    ) -> requests.Response:
        if self.stream_session is None:
            raise Exception("KB not configured")

        response = self.stream_session.post(
            f"{self.url}{ACTIVITY_LOG_QUERY_URL.format(type=type.value)}",
            json=query.model_dump(mode="json", exclude_unset=True),
            stream=True,
        )
        handle_http_sync_errors(response)
        return response

    def logs_download(
        self,
        type: EventType,
        query: Union[
            DownloadActivityLogsQuery,
            DownloadActivityLogsSearchQuery,
            DownloadActivityLogsAskQuery,
        ],
        download_format: DownloadFormat,
    ):
        if self.reader_session is None:
            raise Exception("KB not configured")
        download_request_url = f"{self.url}{ACTIVITY_LOG_URL.format(type=type.value)}"
        format_header_value = DOWNLOAD_FORMAT_HEADERS.get(download_format)
        if format_header_value is None:
            raise ValueError()
        response: httpx.Response = self.reader_session.post(
            download_request_url,
            json=query.model_dump(mode="json", exclude_unset=True),
            headers={"accept": format_header_value},
        )
        handle_http_sync_errors(response)
        return response

    def get_download_request(
        self,
        request_id: str,
    ):
        if self.reader_session is None:
            raise Exception("KB not configured")
        download_request_url = f"{self.url}{ACTIVITY_LOG_DOWNLOAD_REQUEST_URL.format(request_id=request_id)}"
        response: httpx.Response = self.reader_session.get(download_request_url)
        handle_http_sync_errors(response)
        return response

    def remi_query(
        self,
        query: RemiQuery,
    ) -> httpx.Response:
        if self.reader_session is None:
            raise Exception("KB not configured")

        response: httpx.Response = self.reader_session.post(
            f"{self.url}{REMI_QUERY_URL}",
            json=query.model_dump(mode="json", exclude_unset=True),
            timeout=10,
        )
        handle_http_sync_errors(response)
        return response

    def get_remi_event(
        self,
        event_id: int,
    ) -> httpx.Response:
        if self.reader_session is None:
            raise Exception("KB not configured")

        response: httpx.Response = self.reader_session.get(
            f"{self.url}{REMI_EVENT_URL.format(event_id=event_id)}"
        )
        handle_http_sync_errors(response)
        return response

    def get_remi_scores(
        self,
        _from: datetime,
        to: Optional[datetime],
        aggregation: Aggregation,
    ) -> httpx.Response:
        if self.reader_session is None:
            raise Exception("KB not configured")
        params = {"from": _from.isoformat(), "aggregation": aggregation.value}
        if to:
            params["to"] = to.isoformat()
        response: httpx.Response = self.reader_session.get(
            f"{self.url}{REMI_SCORES_URL}", params=params, timeout=10
        )
        handle_http_sync_errors(response)
        return response

    def list_tasks(self) -> httpx.Response:
        if self.reader_session is None:
            raise Exception("KB not configured")

        response: httpx.Response = self.reader_session.get(f"{self.url}{LIST_TASKS}")
        handle_http_sync_errors(response)
        return response

    def start_task(self, body: TaskStartKB) -> httpx.Response:
        if self.writer_session is None:
            raise Exception("KB not configured")

        response: httpx.Response = self.writer_session.post(
            f"{self.url}{START_TASK}",
            json=body.model_dump(mode="json", exclude_unset=True),
        )
        handle_http_sync_errors(response)
        return response

    def delete_task(self, task_id: str, cleanup: bool = False) -> httpx.Response:
        if self.writer_session is None:
            raise Exception("KB not configured")

        params = None
        if cleanup:
            params = {"cleanup": "true"}

        response: httpx.Response = self.writer_session.delete(
            f"{self.url}{DELETE_TASK.format(task_id=task_id)}", params=params
        )
        handle_http_sync_errors(response)
        return response

    def stop_task(self, task_id: str) -> httpx.Response:
        if self.writer_session is None:
            raise Exception("KB not configured")

        response: httpx.Response = self.writer_session.post(
            f"{self.url}{STOP_TASK.format(task_id=task_id)}",
        )
        handle_http_sync_errors(response)
        return response

    def get_task(self, task_id: str) -> httpx.Response:
        if self.reader_session is None:
            raise Exception("KB not configured")

        response: httpx.Response = self.reader_session.get(
            f"{self.url}{GET_TASK.format(task_id=task_id)}",
        )
        handle_http_sync_errors(response)
        return response

    def restart_task(self, task_id: str) -> httpx.Response:
        if self.writer_session is None:
            raise Exception("KB not configured")

        response: httpx.Response = self.writer_session.post(
            f"{self.url}{RESTART_TASK.format(task_id=task_id)}",
        )
        handle_http_sync_errors(response)
        return response

    def list_extract_strategies(self) -> httpx.Response:
        if self.reader_session is None:
            raise Exception("KB not configured")

        response: httpx.Response = self.reader_session.get(
            f"{self.url}{EXTRACT_STRATEGIES}"
        )
        handle_http_sync_errors(response)
        return response

    def add_extract_strategy(self, config: ExtractConfig) -> httpx.Response:
        if self.writer_session is None:
            raise Exception("KB not configured")

        response: httpx.Response = self.writer_session.post(
            f"{self.url}{EXTRACT_STRATEGIES}",
            json=config.model_dump(mode="json", exclude_unset=True),
        )
        handle_http_sync_errors(response)
        return response

    def delete_extract_strategy(self, strategy_id: str) -> httpx.Response:
        if self.writer_session is None:
            raise Exception("KB not configured")

        response: httpx.Response = self.writer_session.delete(
            f"{self.url}{DELETE_EXTRACT_STRATEGY.format(id=strategy_id)}",
        )
        handle_http_sync_errors(response)
        return response

    def list_split_strategies(self) -> httpx.Response:
        if self.reader_session is None:
            raise Exception("KB not configured")

        response: httpx.Response = self.reader_session.get(
            f"{self.url}{SPLIT_STRATEGIES}"
        )
        handle_http_sync_errors(response)
        return response

    def add_split_strategy(self, config: SplitConfiguration) -> httpx.Response:
        if self.writer_session is None:
            raise Exception("KB not configured")

        response: httpx.Response = self.writer_session.post(
            f"{self.url}{SPLIT_STRATEGIES}",
            json=config.model_dump(mode="json", exclude_unset=True),
        )
        handle_http_sync_errors(response)
        return response

    def delete_split_strategy(self, strategy_id: str) -> httpx.Response:
        if self.writer_session is None:
            raise Exception("KB not configured")

        response: httpx.Response = self.writer_session.delete(
            f"{self.url}{DELETE_SPLIT_STRATEGY.format(id=strategy_id)}",
        )
        handle_http_sync_errors(response)
        return response


class AsyncNucliaDBClient(BaseNucliaDBClient):
    reader_session: Optional[httpx.AsyncClient] = None
    writer_session: Optional[httpx.AsyncClient] = None
    ndb: NucliaDBAsync

    def __init__(
        self,
        *,
        environment: Environment = Environment.CLOUD,
        url: Optional[str] = None,
        api_key: Optional[str] = None,
        user_token: Optional[str] = None,
        region: Optional[str] = None,
        base_url: Optional[str] = None,
    ):
        super().__init__(
            environment=environment,
            url=url,
            api_key=api_key,
            user_token=user_token,
            region=region,
            base_url=base_url,
        )

        self.ndb = NucliaDBAsync(
            region=self.region, url=self.base_url, api_key=api_key, headers=self.headers
        )

        if url is not None:
            self.reader_session = build_httpx_async_client(
                headers=self.reader_headers,
                base_url=url,  # type: ignore
            )
            self.writer_session = build_httpx_async_client(
                headers=self.writer_headers,
                base_url=url,  # type: ignore
            )

    async def notifications(self):
        if self.url is None or self.reader_session is None:
            raise Exception("KB not configured")
        url = f"{self.url}{NOTIFICATIONS}"
        req = self.reader_session.build_request("GET", url, timeout=3660)
        response = await self.reader_session.send(req, stream=True)
        await handle_http_async_errors(response)
        return response

    async def ask(
        self,
        request: AskRequest,
        extra_headers: Optional[dict[str, str]] = None,
        timeout: int = 1000,
    ):
        if self.url is None or self.reader_session is None:
            raise Exception("KB not configured")
        url = f"{self.url}{ASK_URL}"
        req = self.reader_session.build_request(
            "POST",
            url,
            json=request.model_dump(),
            headers=extra_headers,
            timeout=timeout,
        )
        response = await self.reader_session.send(req, stream=True)
        await handle_http_async_errors(response)
        return response

    async def download_export(self, export_id: str, path: str, chunk_size: int):
        if self.reader_session is None:
            raise Exception("KB not configured")

        parent = "/".join(path.split(os.sep)[:-1])
        if os.path.exists(parent) is False:
            os.makedirs(parent)

        status = await self.ndb.export_status(kbid=self.kbid, export_id=export_id)
        export_size = status.total

        url = DOWNLOAD_EXPORT_URL.format(export_id=export_id)
        async with self.reader_session.stream("GET", url) as content:
            async with aiofiles.open(path, "wb") as file:
                with tqdm(
                    desc="Downloading data",
                    total=export_size,
                    unit="iB",
                    unit_scale=True,
                ) as pbar:
                    async for chunk in content.aiter_bytes(chunk_size):
                        pbar.update(len(chunk))
                        await file.write(chunk)

    async def download(self, uri: str) -> bytes:
        # uri has format
        # /kb/2a00d5b4-cfcc-48eb-85ac-d70bfd38b26d/resource/41d02aac4ade48098b23e38141807738/file/file/download/field
        # we need to remove the kb url
        if self.reader_session is None:
            raise Exception("KB not configured")

        uri_parts = uri.split("/")
        if len(uri_parts) < 9:
            raise AttributeError("Not a valid download uri")

        new_uri = "/".join(uri_parts[3:])
        url = DOWNLOAD_URL.format(uri=new_uri)
        response = await self.reader_session.get(url)
        await handle_http_async_errors(response)
        return response.content

    @backoff.on_exception(
        backoff.expo,
        RateLimitError,
        jitter=backoff.random_jitter,
        max_tries=5,
        factor=10,
    )
    async def start_tus_upload(
        self,
        size: int,
        filename: str,
        field: Optional[str] = None,
        rid: Optional[str] = None,
        md5: Optional[str] = None,
        content_type: str = "application/octet-stream",
        extract_strategy: Optional[str] = None,
        split_strategy: Optional[str] = None,
    ):
        if self.writer_session is None:
            raise Exception("KB not configured")

        if rid is not None:
            if field is None:
                field = filename
            url = TUS_UPLOAD_RESOURCE_URL.format(rid=rid, field=field)
        else:
            url = TUS_UPLOAD_URL
        encoded_filename = base64.b64encode(filename.encode()).decode()
        headers = {
            "upload-length": str(size),
            "tus-resumable": "1.0.0",
            "upload-metadata": f"filename {encoded_filename}",
            "content-type": content_type,
        }
        if md5 is not None:
            headers["upload-metadata"] += (
                f",md5 {base64.b64encode(md5.encode()).decode()}"
            )
        if extract_strategy is not None:
            headers["x-extract-strategy"] = extract_strategy
        if split_strategy is not None:
            headers["x-split-strategy"] = split_strategy

        response = await self.writer_session.post(url, headers=headers)
        await handle_http_async_errors(response)
        return response.headers.get("Location")

    async def patch_tus_upload(self, upload_url: str, data: bytes, offset: int) -> int:
        if self.writer_session is None:
            raise Exception("KB not configured")

        headers = {
            "upload-offset": str(offset),
        }
        url = httpx.URL(upload_url)
        if url.is_relative_url:
            # Relative path starting with /kb/..., we remove /kb/kbid (the base_url of the session includes it)
            url = httpx.URL("/".join(url.path.split("/")[3:]))

        response = await self.writer_session.patch(url, headers=headers, content=data)
        await handle_http_async_errors(response)
        return int(response.headers.get("Upload-Offset"))

    async def summarize(
        self,
        request: SummarizeRequest,
        extra_headers: Optional[dict[str, str]] = None,
        timeout: int = 1000,
    ):
        if self.url is None or self.writer_session is None:
            raise Exception("KB not configured")
        url = f"{self.url}{SUMMARIZE_URL}"
        assert self.reader_session
        response = await self.reader_session.post(
            url, json=request.model_dump(), headers=extra_headers, timeout=timeout
        )
        await handle_http_async_errors(response)
        return response

    async def logs_query(
        self,
        type: EventType,
        query: Union[ActivityLogsQuery, ActivityLogsSearchQuery, ActivityLogsAskQuery],
    ) -> httpx.Response:
        if self.reader_session is None:
            raise Exception("KB not configured")

        response = await self.reader_session.post(
            f"{self.url}{ACTIVITY_LOG_QUERY_URL.format(type=type.value)}",
            json=query.model_dump(mode="json", exclude_unset=True),
        )
        await handle_http_async_errors(response)
        return response

    async def logs_download(
        self,
        type: EventType,
        query: Union[
            DownloadActivityLogsQuery,
            DownloadActivityLogsSearchQuery,
            DownloadActivityLogsAskQuery,
        ],
        download_format: DownloadFormat,
    ):
        if self.reader_session is None:
            raise Exception("KB not configured")
        download_request_url = f"{self.url}{ACTIVITY_LOG_URL.format(type=type.value)}"
        format_header_value = DOWNLOAD_FORMAT_HEADERS.get(download_format)
        if format_header_value is None:
            raise ValueError()
        response: httpx.Response = await self.reader_session.post(
            download_request_url,
            json=query.model_dump(mode="json", exclude_unset=True),
            headers={"accept": format_header_value},
        )
        await handle_http_async_errors(response)
        return response

    async def get_download_request(
        self,
        request_id: str,
    ):
        if self.reader_session is None:
            raise Exception("KB not configured")
        download_request_url = f"{self.url}{ACTIVITY_LOG_DOWNLOAD_REQUEST_URL.format(request_id=request_id)}"
        response: httpx.Response = await self.reader_session.get(download_request_url)
        await handle_http_async_errors(response)
        return response

    async def remi_query(
        self,
        query: RemiQuery,
    ) -> httpx.Response:
        if self.reader_session is None:
            raise Exception("KB not configured")

        response: httpx.Response = await self.reader_session.post(
            f"{self.url}{REMI_QUERY_URL}",
            json=query.model_dump(mode="json", exclude_unset=True),
            timeout=10,
        )
        await handle_http_async_errors(response)
        return response

    async def get_remi_event(
        self,
        event_id: int,
    ) -> httpx.Response:
        if self.reader_session is None:
            raise Exception("KB not configured")

        response: httpx.Response = await self.reader_session.get(
            f"{self.url}{REMI_EVENT_URL.format(event_id=event_id)}"
        )
        await handle_http_async_errors(response)
        return response

    async def get_remi_scores(
        self,
        _from: datetime,
        to: Optional[datetime],
        aggregation: Aggregation,
    ) -> httpx.Response:
        if self.reader_session is None:
            raise Exception("KB not configured")
        params = {"from": _from.isoformat(), "aggregation": aggregation.value}
        if to:
            params["to"] = to.isoformat()
        response: httpx.Response = await self.reader_session.get(
            f"{self.url}{REMI_SCORES_URL}", params=params, timeout=10
        )
        await handle_http_async_errors(response)
        return response

    async def list_tasks(self) -> httpx.Response:
        if self.reader_session is None:
            raise Exception("KB not configured")

        response: httpx.Response = await self.reader_session.get(
            f"{self.url}{LIST_TASKS}"
        )
        await handle_http_async_errors(response)
        return response

    async def start_task(self, body: TaskStartKB) -> httpx.Response:
        if self.writer_session is None:
            raise Exception("KB not configured")

        response: httpx.Response = await self.writer_session.post(
            f"{self.url}{START_TASK}",
            json=body.model_dump(mode="json", exclude_unset=True),
        )
        await handle_http_async_errors(response)
        return response

    async def delete_task(self, task_id: str, cleanup: bool = False) -> httpx.Response:
        if self.writer_session is None:
            raise Exception("KB not configured")

        params = None
        if cleanup:
            params = {"cleanup": "true"}

        response: httpx.Response = await self.writer_session.delete(
            f"{self.url}{DELETE_TASK.format(task_id=task_id)}", params=params
        )
        await handle_http_async_errors(response)
        return response

    async def stop_task(self, task_id: str) -> httpx.Response:
        if self.writer_session is None:
            raise Exception("KB not configured")

        response: httpx.Response = await self.writer_session.post(
            f"{self.url}{STOP_TASK.format(task_id=task_id)}",
        )
        await handle_http_async_errors(response)
        return response

    async def get_task(self, task_id: str) -> httpx.Response:
        if self.reader_session is None:
            raise Exception("KB not configured")

        response: httpx.Response = await self.reader_session.get(
            f"{self.url}{GET_TASK.format(task_id=task_id)}",
        )
        await handle_http_async_errors(response)
        return response

    async def restart_task(self, task_id: str) -> httpx.Response:
        if self.writer_session is None:
            raise Exception("KB not configured")

        response: httpx.Response = await self.writer_session.post(
            f"{self.url}{RESTART_TASK.format(task_id=task_id)}",
        )
        await handle_http_async_errors(response)
        return response

    async def list_extract_strategies(self) -> httpx.Response:
        if self.reader_session is None:
            raise Exception("KB not configured")

        response: httpx.Response = await self.reader_session.get(
            f"{self.url}{EXTRACT_STRATEGIES}"
        )
        await handle_http_async_errors(response)
        return response

    async def add_extract_strategy(self, config: ExtractConfig) -> httpx.Response:
        if self.writer_session is None:
            raise Exception("KB not configured")

        response: httpx.Response = await self.writer_session.post(
            f"{self.url}{EXTRACT_STRATEGIES}",
            json=config.model_dump(mode="json", exclude_unset=True),
        )
        await handle_http_async_errors(response)
        return response

    async def delete_extract_strategy(self, strategy_id: str) -> httpx.Response:
        if self.writer_session is None:
            raise Exception("KB not configured")

        response: httpx.Response = await self.writer_session.delete(
            f"{self.url}{DELETE_EXTRACT_STRATEGY.format(id=strategy_id)}",
        )
        await handle_http_async_errors(response)
        return response

    async def list_split_strategies(self) -> httpx.Response:
        if self.reader_session is None:
            raise Exception("KB not configured")

        response: httpx.Response = await self.reader_session.get(
            f"{self.url}{SPLIT_STRATEGIES}"
        )
        await handle_http_async_errors(response)
        return response

    async def add_split_strategy(self, config: SplitConfiguration) -> httpx.Response:
        if self.writer_session is None:
            raise Exception("KB not configured")

        response: httpx.Response = await self.writer_session.post(
            f"{self.url}{SPLIT_STRATEGIES}",
            json=config.model_dump(mode="json", exclude_unset=True),
        )
        await handle_http_async_errors(response)
        return response

    async def delete_split_strategy(self, strategy_id: str) -> httpx.Response:
        if self.writer_session is None:
            raise Exception("KB not configured")

        response: httpx.Response = await self.writer_session.delete(
            f"{self.url}{DELETE_SPLIT_STRATEGY.format(id=strategy_id)}",
        )
        await handle_http_async_errors(response)
        return response



import asyncio
import base64
import datetime
import json
import webbrowser
from time import time
from typing import Any, Dict, List, Optional, Tuple

from httpx import AsyncClient, Client, ConnectError
from nucliadb_models.resource import KnowledgeBoxObj
from prompt_toolkit import prompt
from pydantic import TypeAdapter
from tabulate import tabulate

from nuclia import get_global_url, get_regional_url, is_nuclia_hosted
from nuclia.config import (
    Account,
    Config,
    EphemeralToken,
    KnowledgeBox,
    PersonalTokenCreate,
    PersonalTokenItem,
    User,
    Zone,
    retrieve_account,
    retrieve_nua,
)
from nuclia.exceptions import NeedUserToken, UserTokenExpired
from nuclia.lib.utils import build_httpx_async_client, build_httpx_client
from nuclia.sdk.logger import logger

USER = "/api/v1/user/welcome"
MEMBER = "/api/v1/user"
ACCOUNTS = "/api/v1/accounts"
ZONES = "/api/v1/zones"
LIST_KBS = "/api/v1/account/{account}/kbs"
LIST_AGENTS = "/api/v1/account/{account}/kbs?mode=agent"
VERIFY_NUA = "/api/authorizer/info"
PERSONAL_TOKENS = "/api/v1/user/pa_tokens"
PERSONAL_TOKEN = "/api/v1/user/pa_token/{token_id}"
SA_EPHEMERAL_TOKEN = "/api/v1/ephemeral_token"


class BaseNucliaAuth:
    _inner_config: Optional[Config] = None

    @property
    def _config(self) -> Config:
        if self._inner_config is None:
            from nuclia.data import get_config

            self._inner_config = get_config()
        return self._inner_config

    def get_account_id(self, account_slug: str) -> str:
        account_obj = retrieve_account(self._config.accounts or [], account_slug)
        if not account_obj:
            raise ValueError(f"Account {account_slug} not found")
        return account_obj.id

    def list_nuas(self):
        result = []
        result.extend(
            self._config.nuas_token if self._config.nuas_token is not None else []
        )
        return result

    def default_nua(self, nua: str):
        nuas = self._config.nuas_token if self._config.nuas_token is not None else []
        nua_obj = retrieve_nua(nuas, nua)

        if nua_obj is None:
            raise KeyError("NUA KEY not found")
        self._config.set_default_nua(nua_obj.client_id)
        self._config.save()

    def default_kb(self, kbid: str):
        self._config.set_default_kb(kbid=kbid)

    def unset_kb(self, kbid: str):
        self._config.unset_default_kb(kbid=kbid)


def print_config(config: Config):
    if config.default:
        if config.default.account:
            print(f"Default account:        {config.default.account}")
        if config.default.kbid:
            kb_obj = config.get_kb(config.default.kbid)
            print(f"Default KnowledgeBox:  {kb_obj}\n")
        if config.default.agent_id:
            agent_obj = config.get_agent(config.default.agent_id)
            print(f"Default Retrieval Agent Orchestrator:           {agent_obj}\n")


def print_nuas(config: Config):
    if config.default:
        if config.default.nua:
            nua = config.get_nua(config.default.nua)
            print(
                f"Default NUA KEY:        {nua.region} {nua.client_id} ({nua.account}) \n"
            )

    if len(config.nuas_token):
        print("NUA Keys Registered locally: \n")
        data = []
        for nua in config.nuas_token:
            data.append(
                [
                    nua.client_id,
                    nua.region,
                    nua.account,
                ]
            )
        print(
            tabulate(
                data,
                headers=[
                    "Client ID",
                    "Region",
                    "Account",
                ],
            )
        )
        print()


class NucliaAuth(BaseNucliaAuth):
    client: Client

    def __init__(self):
        self.client = build_httpx_client()

    def show(self) -> None:
        self._show_user()
        print_config(self._config)

        data_kbs = []
        data_agents = []
        for account in self.accounts():
            for kb in self.kbs(account.id):
                try:
                    kb_obj = self._config.get_kb(kb.id)
                except StopIteration:
                    kb_obj = None

                role: Optional[str] = ""
                if (
                    kb_obj is not None
                    and kb_obj.region is not None
                    and kb_obj.token is not None
                ):
                    permissions = self.client.get(
                        get_regional_url(kb_obj.region, "/api/authorizer/info"),
                        headers={"X-NUCLIA-SERVICEACCOUNT": f"Bearer {kb_obj.token}"},
                    )
                    if permissions.status_code == 200:
                        role = permissions.json().get("user", {}).get("role")
                        if role is not None:
                            role = role.lstrip("S")

                default = ""
                if (
                    self._config.default is not None
                    and self._config.default.kbid == kb.id
                ):
                    default = "*"

                data_kbs.append(
                    [
                        default,
                        account.slug,
                        kb.id,
                        kb.slug,
                        kb.title,
                        kb.region,
                        kb.url,
                        role,
                    ]
                )
            for agent in self.agents(account.id):
                try:
                    agent_obj = self._config.get_agent(agent.id)
                except StopIteration:
                    agent_obj = None

                arole: Optional[str] = ""
                if (
                    agent_obj is not None
                    and agent_obj.region is not None
                    and agent_obj.token is not None
                ):
                    permissions = self.client.get(
                        get_regional_url(agent_obj.region, "/api/authorizer/info"),
                        headers={
                            "X-NUCLIA-SERVICEACCOUNT": f"Bearer {agent_obj.token}"
                        },
                    )
                    if permissions.status_code == 200:
                        arole = permissions.json().get("user", {}).get("role")
                        if arole is not None:
                            arole = arole.lstrip("S")

                default = ""
                if (
                    self._config.default is not None
                    and self._config.default.agent_id == agent.id
                ):
                    default = "*"

                data_agents.append(
                    [
                        default,
                        account.slug,
                        agent.id,
                        agent.slug,
                        agent.title,
                        agent.region,
                        agent.url,
                        arole,
                    ]
                )

        print(
            tabulate(
                data_kbs,
                headers=[
                    "Default",
                    "Account ID",
                    "KnowledgeBox ID",
                    "KnowledgeBox slug",
                    "KnowledgeBox Title",
                    "KnowledgeBox Region",
                    "KnowledgeBox URL",
                    "Local Token Available",
                ],
            )
        )
        print()
        print(
            tabulate(
                data_agents,
                headers=[
                    "Default",
                    "Account ID",
                    "Retrieval Agent ID",
                    "Retrieval Agent slug",
                    "Retrieval Agent Title",
                    "Retrieval Agent Region",
                    "Retrieval Agent URL",
                    "Local Token Available",
                ],
            )
        )
        print()

    def nucliadb(self, url: str = "http://localhost:8080"):
        """
        Setup a local NucliaDB. Needs to be the base url of the NucliaDB server
        """
        resp = self.client.get(url)
        if resp.status_code != 200 or b"Nuclia" not in resp.content:
            raise Exception("Not a valid URL")
        self._config.set_default_nucliadb(nucliadb=url)

    def kb(self, url: str, token: Optional[str] = None) -> Optional[str]:
        url = url.strip("/")
        kb = self.validate_kb(url, token)
        if kb:
            logger.info("Validated")
            self._config.set_kb_token(
                url=url,
                token=token,
                title=kb.config.title if kb.config is not None else "",
                kbid=kb.uuid,
            )
            self._config.set_default_kb(kbid=kb.uuid)
        else:
            logger.error("Invalid service token")
        return kb.uuid if kb is not None else None

    def agent(self, url: str, token: Optional[str] = None) -> Optional[str]:
        url = url.strip("/")
        agent = self.validate_kb(url, token)
        if agent:
            logger.info("Validated")
            # Extract account ID if using service token
            account_id = None
            if token is not None:
                from nuclia.config import extract_region

                region = extract_region(url)
                if region:
                    account_id, _, _ = self.get_account_from_service_token(
                        region, token
                    )

            self._config.set_agent_token(
                url=url,
                token=token,
                title=agent.config.title if agent.config is not None else "",
                agent_id=agent.uuid,
                account_id=account_id,
            )
            self._config.set_default_agent(agent_id=agent.uuid)
        else:
            logger.error("Invalid service token")
        return None if agent is None else agent.uuid

    def nua(self, token: str) -> Optional[str]:
        client_id, account_type, account, base_region = self.validate_nua(token)
        if account is not None and client_id is not None and base_region is not None:
            logger.info("Validated")
            self._config.set_nua_token(
                client_id=client_id,
                account_type=account_type,
                account=account,
                base_region=base_region,
                token=token,
            )
            self.default_nua(client_id)
            return client_id
        else:
            logger.error("Invalid NUA token")
            return None

    def nuas(self):
        print_nuas(self._config)

    def validate_nua(
        self, token: str
    ) -> Tuple[Optional[str], Optional[str], Optional[str], Optional[str]]:
        # Validate the code is ok
        token_payload = token.split(".")[1]
        token_payload_decoded = str(base64.b64decode(token_payload + "=="), "utf-8")
        payload = json.loads(token_payload_decoded)
        base_path = payload["iss"]
        url = base_path.strip("/") + VERIFY_NUA
        resp = self.client.get(
            url,
            headers={"x-nuclia-nuakey": f"Bearer {token}"},
        )
        if resp.status_code == 200:
            data = resp.json().get("user")
            return (
                data.get("user_id"),
                data.get("account_type"),
                data.get("account_id"),
                payload.get("iss"),
            )
        else:
            return None, None, None, None

    def get_account_from_service_token(
        self, region: str, token: str
    ) -> Tuple[Optional[str], Optional[str], Optional[str]]:
        """
        Extract account ID, KB/agent ID, and role from a service account token (API key).

        Args:
            region: The region where the KB/agent is located (e.g., 'europe-1')
            token: The service account token (API key)

        Returns:
            Tuple of (account_id, kb_id, role) or (None, None, None) if invalid
        """
        url = get_regional_url(region, "/api/authorizer/info")
        resp = self.client.get(
            url,
            headers={"x-nuclia-serviceaccount": f"Bearer {token}"},
        )
        if resp.status_code == 200:
            data = resp.json().get("user")
            if data:
                return (
                    data.get("account_id"),
                    data.get("kb_id"),
                    data.get("role"),
                )
        return None, None, None

    def validate_kb(
        self, url: str, token: Optional[str] = None
    ) -> Optional[KnowledgeBoxObj]:
        # Validate the code is ok
        data = None
        if token is None:
            if is_nuclia_hosted(url):
                # Use nuclia User Auth
                data = self._request("GET", url)
            else:
                # Validate OSS version
                resp = self.client.get(
                    url,
                    headers={"X-NucliaDB-ROLES": "READER"},
                )
                if resp.status_code == 200:
                    data = resp.json()
        else:
            # Validate Cloud version
            resp = self.client.get(
                url,
                headers={"X-Nuclia-Serviceaccount": f"Bearer {token}"},
            )
            if resp.status_code == 200:
                data = resp.json()
        if data is not None:
            return KnowledgeBoxObj.model_validate(data)
        else:
            return None

    def get_user(self) -> User:
        resp = self._request("GET", get_global_url(MEMBER))
        assert resp
        return User.model_validate_json(resp.json)

    def _show_user(self):
        resp = None
        try:
            resp = self._request("GET", get_global_url(MEMBER))
        except NeedUserToken:
            print("No user logged in.")
        if resp:
            print()
            print(f"User: {resp.get('name')} <{resp.get('email')}>")
            print(f"Type: {resp.get('type')}")
            print()

    def login(self):
        """
        Lets redirect the user to the UI to capture a token
        """
        if self._validate_user_token():
            logger.info("Logged in!")
            self._show_user()
            return

        webbrowser.open(get_global_url("/redirect?display=token"))
        # we cannot use Python's `input` here because the copy/pasted token is too long
        code = prompt("Follow the browser flow and copy the token and paste it here:")
        print("Checking...")
        self.set_user_token(code)

    def logout(self):
        """
        Remove the current user token.
        """
        self._config.remove_user_token()

    def set_user_token(self, code: str):
        if self._validate_user_token(code):
            self._config.set_user_token(code)
            logger.info("Auth completed!")
            self.post_login()
        else:
            print("Invalid token auth not completed")

    def _extract_account(self, token: str) -> str:
        base64url = token.split(".")[1]
        data = json.loads(
            base64.urlsafe_b64decode(base64url + "=" * (4 - len(base64url) % 4))
        )
        return data.get("jti")

    def _validate_user_token(self, code: Optional[str] = None) -> bool:
        # Validate the code is ok
        if code is None:
            code = self._config.token
        resp = self.client.get(
            get_global_url(USER),
            headers={"Authorization": f"Bearer {code}"},
        )
        if resp.status_code == 200:
            return True
        else:
            return False

    def post_login(self):
        self.accounts()
        self.zones()

    def create_ephemeral_token(
        self, kbid: str, ttl: Optional[int] = None
    ) -> EphemeralToken:
        # Try to get as KB first, then as agent
        kb_obj = None
        try:
            kb_obj = self._config.get_kb(kbid)
        except (StopIteration, KeyError):
            try:
                kb_obj = self._config.get_agent(kbid)
            except (StopIteration, KeyError):
                pass

        if kb_obj is None:
            raise ValueError("KnowledgeBox or Agent not found")

        if kb_obj.region is None:
            raise ValueError("KnowledgeBox/Agent region not set")

        payload = {}
        if ttl is not None:
            payload["ttl"] = ttl
        resp = self.client.post(
            get_regional_url(kb_obj.region, SA_EPHEMERAL_TOKEN),
            headers={"X-NUCLIA-SERVICEACCOUNT": f"Bearer {kb_obj.token}"},
            json=payload,
        )
        return EphemeralToken(token=resp.json().get("token"))

    def create_personal_token(
        self, description: str, days: int = 90, login: bool = False
    ) -> PersonalTokenCreate:
        expiration_date = datetime.datetime.now() + datetime.timedelta(days=days)
        resp = self._request(
            "POST",
            get_global_url(PERSONAL_TOKENS),
            {
                "description": description,
                "expiration_date": expiration_date.isoformat(),
            },
        )
        token = PersonalTokenCreate.model_validate(resp)
        if login:
            self.set_user_token(token.token)
        return token

    def delete_personal_token(self, token_id: str) -> None:
        self._request(
            "DELETE", get_global_url(PERSONAL_TOKEN.format(token_id=token_id))
        )

    def list_personal_tokens(self) -> List[PersonalTokenItem]:
        resp = self._request("GET", get_global_url(PERSONAL_TOKENS))
        ta = TypeAdapter(List[PersonalTokenItem])
        return ta.validate_python(resp)

    def _request(
        self, method: str, path: str, data: Optional[Any] = None, remove_null=True
    ):
        if not self._config.token:
            raise NeedUserToken()
        kwargs: Dict[str, Any] = {
            "headers": {"Authorization": f"Bearer {self._config.token}"}
        }
        if data is not None:
            if remove_null:
                data = {k: v for k, v in data.items() if v is not None}
            kwargs["content"] = json.dumps(data)

        resp = self.client.request(
            method,
            path,
            **kwargs,
        )
        if resp.status_code == 204:
            return None
        elif resp.status_code >= 200 and resp.status_code < 300:
            return resp.json()
        elif resp.status_code >= 300 and resp.status_code < 400:
            return None
        elif resp.status_code == 403 or resp.status_code == 401:
            raise UserTokenExpired()
        else:
            raise Exception({"status": resp.status_code, "message": resp.text})

    def accounts(self) -> List[Account]:
        accounts = self._request("GET", get_global_url(ACCOUNTS))
        result: List[Account] = []
        self._config.accounts = []
        if accounts is None:
            return result
        for account in accounts:
            account_obj = Account.model_validate(account)
            result.append(account_obj)
            self._config.accounts.append(account_obj)
        self._config.save()
        return result

    def zones(self) -> List[Zone]:
        zones = self._request("GET", get_global_url(ZONES))
        if self._config.accounts is None:
            self._config.accounts = []
        self._config.zones = []
        result: List[Zone] = []
        if zones is None:
            return result
        for zone in zones:
            zone_obj = Zone.model_validate(zone)
            result.append(zone_obj)
            self._config.zones.append(zone_obj)
        self._config.save()
        return result

    def kbs(self, account: str) -> List[KnowledgeBox]:
        result: List[KnowledgeBox] = []
        zones = self.zones()
        for zoneObj in zones:
            zoneSlug = zoneObj.slug
            if not zoneSlug:
                continue
            path = get_regional_url(zoneSlug, LIST_KBS.format(account=account))
            try:
                kbs = self._request("GET", path)
            except UserTokenExpired:
                return []
            except ConnectError:
                logger.error(
                    f"Connection error to {get_regional_url(zoneSlug, '')}, skipping zone"
                )
                continue
            if kbs is not None:
                for kb in kbs:
                    url = get_regional_url(zoneSlug, f"/api/v1/kb/{kb['id']}")
                    kb_obj = KnowledgeBox(
                        url=url,
                        id=kb["id"],
                        slug=kb["slug"],
                        title=kb["title"],
                        account=account,
                        region=zoneSlug,
                    )
                    result.append(kb_obj)
        return result

    def agents(self, account: str) -> List[KnowledgeBox]:
        result: List[KnowledgeBox] = []
        zones = self.zones()
        for zoneObj in zones:
            zoneSlug = zoneObj.slug
            if not zoneSlug:
                continue
            path = get_regional_url(zoneSlug, LIST_AGENTS.format(account=account))
            try:
                agents = self._request("GET", path)
            except UserTokenExpired:
                return []
            except ConnectError:
                logger.error(
                    f"Connection error to {get_regional_url(zoneSlug, '')}, skipping zone"
                )
                continue
            if agents is not None:
                for agent in agents:
                    url = get_regional_url(zoneSlug, f"/api/v1/agent/{agent['id']}")
                    ra_obj = KnowledgeBox(
                        url=url,
                        id=agent["id"],
                        slug=agent["slug"],
                        title=agent["title"],
                        account=account,
                        region=zoneSlug,
                    )
                    result.append(ra_obj)
        return result


class AsyncNucliaAuth(BaseNucliaAuth):
    client: AsyncClient

    def __init__(self):
        self.client = build_httpx_async_client()
        self._cache = {}  # Manual cache storage
        self._lock = asyncio.Lock()

    async def show(self):
        await self._show_user()
        print_config(self._config)

    async def nucliadb(self, url: str = "http://localhost:8080"):
        """
        Setup a local NucliaDB. Needs to be the base url of the NucliaDB server
        """
        resp = await self.client.get(url)
        if resp.status_code != 200 or b"Nuclia" not in resp.content:
            raise Exception("Not a valid URL")
        self._config.set_default_nucliadb(nucliadb=url)

    async def kb(self, url: str, token: Optional[str] = None) -> Optional[str]:
        url = url.strip("/")
        kb = await self.validate_kb(url, token)
        if kb:
            logger.info("Validated")
            self._config.set_kb_token(
                url=url,
                token=token,
                title=kb.config.title if kb.config is not None else "",
                kbid=kb.uuid,
            )
            self._config.set_default_kb(kbid=kb.uuid)
        else:
            logger.error("Invalid service token")
        return kb.uuid if kb is not None else None

    async def agent(self, url: str, token: Optional[str] = None) -> Optional[str]:
        url = url.strip("/")
        agent = await self.validate_kb(url, token)
        if agent:
            logger.info("Validated")
            # Extract account ID if using service token
            account_id = None
            if token is not None:
                from nuclia.config import extract_region

                region = extract_region(url)
                if region:
                    account_id, _, _ = await self.get_account_from_service_token(
                        region, token
                    )

            self._config.set_agent_token(
                url=url,
                token=token,
                title=agent.config.title if agent.config is not None else "",
                agent_id=agent.uuid,
                account_id=account_id,
            )
            self._config.set_default_agent(agent_id=agent.uuid)
        else:
            logger.error("Invalid service token")
        return None if agent is None else agent.uuid

    async def nua(self, token: str) -> Optional[str]:
        client_id, account_type, account, base_region = await self.validate_nua(token)
        if account is not None and client_id is not None and base_region is not None:
            logger.info("Validated")
            self._config.set_nua_token(
                client_id=client_id,
                account_type=account_type,
                account=account,
                base_region=base_region,
                token=token,
            )
            return client_id
        else:
            logger.error("Invalid NUA token")
            return None

    async def validate_nua(
        self, token: str
    ) -> Tuple[Optional[str], Optional[str], Optional[str], Optional[str]]:
        # Validate the code is ok
        token_payload = token.split(".")[1]
        token_payload_decoded = str(base64.b64decode(token_payload + "=="), "utf-8")
        payload = json.loads(token_payload_decoded)
        base_path = payload["iss"]
        url = base_path.strip("/") + VERIFY_NUA
        resp = await self.client.get(
            url,
            headers={"x-nuclia-nuakey": f"Bearer {token}"},
        )
        if resp.status_code == 200:
            data = resp.json().get("user")
            return (
                data.get("user_id"),
                data.get("account_type"),
                data.get("account_id"),
                payload.get("iss"),
            )
        else:
            return None, None, None, None

    async def get_account_from_service_token(
        self, region: str, token: str
    ) -> Tuple[Optional[str], Optional[str], Optional[str]]:
        """
        Extract account ID, KB/agent ID, and role from a service account token (API key).

        Args:
            region: The region where the KB/agent is located (e.g., 'europe-1')
            token: The service account token (API key)

        Returns:
            Tuple of (account_id, kb_id, role) or (None, None, None) if invalid
        """
        url = get_regional_url(region, "/api/authorizer/info")
        resp = await self.client.get(
            url,
            headers={"x-nuclia-serviceaccount": f"Bearer {token}"},
        )
        if resp.status_code == 200:
            data = resp.json().get("user")
            if data:
                return (
                    data.get("account_id"),
                    data.get("kb_id"),
                    data.get("role"),
                )
        return None, None, None

    async def validate_kb(
        self, url: str, token: Optional[str] = None
    ) -> Optional[KnowledgeBoxObj]:
        # Validate the code is ok
        data = None
        if token is None:
            if is_nuclia_hosted(url):
                # Use nuclia User Auth
                data = await self._request("GET", url)
            else:
                # Validate OSS version
                resp = await self.client.get(
                    url,
                    headers={"X-NucliaDB-ROLES": "READER"},
                )
                if resp.status_code == 200:
                    data = resp.json()
        else:
            # Validate Cloud version
            resp = await self.client.get(
                url,
                headers={"X-Nuclia-Serviceaccount": f"Bearer {token}"},
            )
            if resp.status_code == 200:
                data = resp.json()
        if data is not None:
            return KnowledgeBoxObj.model_validate(data)
        else:
            return None

    async def _show_user(self):
        resp = await self._request("GET", get_global_url(MEMBER))
        assert resp
        print(f"User: {resp.get('name')} <{resp.get('email')}>")
        print(f"Type: {resp.get('type')}")

    async def login(self):
        """
        Lets redirect the user to the UI to capture a token
        """
        if await self._validate_user_token():
            logger.info("Logged in!")
            await self._show_user()
            return

        webbrowser.open(get_global_url("/redirect?display=token"))
        # we cannot use Python's `input` here because the copy/pasted token is too long
        code = prompt("Follow the browser flow and copy the token and paste it here:")
        logger.info("Checking...")
        await self.set_user_token(code)

    def logout(self):
        """
        Remove the current user token.
        """
        self._config.remove_user_token()

    async def set_user_token(self, code: str):
        if await self._validate_user_token(code):
            self._config.set_user_token(code)
            logger.info("Auth completed!")
            await self.post_login()
        else:
            logger.error("Invalid token auth not completed")

    def _extract_account(self, token: str) -> str:
        base64url = token.split(".")[1]
        data = json.loads(
            base64.urlsafe_b64decode(base64url + "=" * (4 - len(base64url) % 4))
        )
        return data.get("jti")

    async def _validate_user_token(self, code: Optional[str] = None) -> bool:
        # Validate the code is ok
        if code is None:
            code = self._config.token
        resp = await self.client.get(
            get_global_url(USER),
            headers={"Authorization": f"Bearer {code}"},
        )
        if resp.status_code == 200:
            return True
        else:
            return False

    async def post_login(self):
        await self.accounts()
        await self.zones()

    async def _cached_request(self, method: str, path: str) -> Any:
        async with self._lock:
            if path in self._cache:
                exp, value = self._cache[(path)]
                if time() <= exp:
                    return value  # Return cached response
            response = await self._request(method, path)
            self._cache[path] = (time() + 60, response)
            return response

    async def _request(
        self, method: str, path: str, data: Optional[Any] = None, remove_null=True
    ):
        if not self._config.token:
            raise NeedUserToken()
        kwargs: Dict[str, Any] = {
            "headers": {"Authorization": f"Bearer {self._config.token}"}
        }

        if data is not None:
            if remove_null:
                data = {k: v for k, v in data.items() if v is not None}
            kwargs["data"] = json.dumps(data)
        resp = await self.client.request(
            method,
            path,
            **kwargs,
        )
        if resp.status_code == 204:
            return None
        elif resp.status_code >= 200 and resp.status_code < 300:
            return resp.json()
        elif resp.status_code >= 300 and resp.status_code < 400:
            return None
        elif resp.status_code == 403 or resp.status_code == 401:
            raise UserTokenExpired()
        else:
            raise Exception({"status": resp.status_code, "message": resp.text})

    async def accounts(self, cached: bool = True) -> List[Account]:
        _request = self._cached_request if cached else self._request
        accounts = await _request("GET", get_global_url(ACCOUNTS))
        result: List[Account] = []
        self._config.accounts = []
        if accounts is None:
            return result
        for account in accounts:
            account_obj = Account.model_validate(account)
            result.append(account_obj)
            self._config.accounts.append(account_obj)
        self._config.save()
        return result

    async def zones(self, cached: bool = True) -> List[Zone]:
        _request = self._cached_request if cached else self._request
        zones = await _request("GET", get_global_url(ZONES))
        if self._config.accounts is None:
            self._config.accounts = []
        self._config.zones = []
        result: List[Zone] = []
        if zones is None:
            return result
        for zone in zones:
            zone_obj = Zone.model_validate(zone)
            result.append(zone_obj)
            self._config.zones.append(zone_obj)
        self._config.save()
        return result

    async def kbs(self, account: str, cached: bool = True) -> List[KnowledgeBox]:
        _request = self._cached_request if cached else self._request
        result: List[KnowledgeBox] = []
        zones = await self.zones()
        for zoneObj in zones:
            zoneSlug = zoneObj.slug
            if not zoneSlug:
                continue
            path = get_regional_url(zoneSlug, LIST_KBS.format(account=account))
            try:
                kbs = await _request("GET", path)
            except UserTokenExpired:
                return result
            except ConnectError:
                logger.error(
                    f"Connection error to {get_regional_url(zoneSlug, '')}, skipping zone"
                )
                continue
            if kbs is not None:
                for kb in kbs:
                    url = get_regional_url(zoneSlug, f"/api/v1/kb/{kb['id']}")
                    kb_obj = KnowledgeBox(
                        url=url,
                        id=kb["id"],
                        slug=kb["slug"],
                        title=kb["title"],
                        account=account,
                        region=zoneSlug,
                    )
                    result.append(kb_obj)
        return result

    async def create_ephemeral_token(
        self, kbid: str, ttl: Optional[int] = None
    ) -> EphemeralToken:
        # Try to get as KB first, then as agent
        kb_obj = None
        try:
            kb_obj = self._config.get_kb(kbid)
        except (StopIteration, KeyError):
            try:
                kb_obj = self._config.get_agent(kbid)
            except (StopIteration, KeyError):
                pass

        if kb_obj is None:
            raise ValueError("KnowledgeBox or Agent not found")

        if kb_obj.region is None:
            raise ValueError("KnowledgeBox/Agent region not set")

        payload = {}
        if ttl is not None:
            payload["ttl"] = ttl
        resp = await self.client.post(
            get_regional_url(kb_obj.region, SA_EPHEMERAL_TOKEN),
            headers={"X-NUCLIA-SERVICEACCOUNT": f"Bearer {kb_obj.token}"},
            json=payload,
        )
        return EphemeralToken(token=resp.json().get("token"))

    async def agents(self, account: str, cached: bool = True) -> List[KnowledgeBox]:
        _request = self._cached_request if cached else self._request
        result: List[KnowledgeBox] = []
        zones = await self.zones()
        for zoneObj in zones:
            zoneSlug = zoneObj.slug
            if not zoneSlug:
                continue
            path = get_regional_url(zoneSlug, LIST_AGENTS.format(account=account))
            try:
                agents = await _request("GET", path)
            except UserTokenExpired:
                return []
            except ConnectError:
                logger.error(
                    f"Connection error to {get_regional_url(zoneSlug, '')}, skipping zone"
                )
                continue
            if agents is not None:
                for agent in agents:
                    url = get_regional_url(zoneSlug, f"/api/v1/agent/{agent['id']}")
                    ra_obj = KnowledgeBox(
                        url=url,
                        id=agent["id"],
                        slug=agent["slug"],
                        title=agent["title"],
                        account=account,
                        region=zoneSlug,
                    )
                    result.append(ra_obj)
        return result



import importlib.metadata
import json
from typing import Optional

import httpx
import requests
from httpx import HTTPStatusError
from httpx import Response as HttpxResponse
from nuclia_models.agent.interaction import AnswerOperation, AragAnswer
from nuclia_models.worker.tasks import TaskDefinition, TaskList
from nucliadb_models.resource import KnowledgeBoxList, ResourceList
from nucliadb_models.search import SyncAskResponse
from requests import HTTPError as RequestsHTTPError
from requests import Response as RequestsResponse
from tabulate import tabulate

from nuclia.exceptions import (
    DuplicateError,
    InvalidPayload,
    RateLimitError,
    UserTokenExpired,
)
from nuclia.lib.models import ActivityLogsOutput

USER_AGENT = f"nuclia.py/{importlib.metadata.version('nuclia')}"

MAX_TITLE_LEN = 50


def handle_http_sync_errors(response):
    try:
        content = response.text
    except (httpx.ResponseNotRead, requests.exceptions.RequestException):
        content = "<streaming content not read>"
    except Exception as e:
        content = f"<error decoding content: {e}>"

    _raise_for_status(response.status_code, content, response=response)


async def handle_http_async_errors(response: httpx.Response):
    try:
        if not response.is_closed and not response.is_stream_consumed:
            await response.aread()
    except Exception as e:
        content = f"<failed to read stream: {e}>"
        _raise_for_status(
            response.status_code, content, response=response, request=response.request
        )
        return  # Defensive
    try:
        content = response.text
    except httpx.ResponseNotRead:
        content = "<streaming content not read>"
    except Exception as e:
        content = f"<error decoding content: {e}>"

    _raise_for_status(
        response.status_code, content, response=response, request=response.request
    )


def _raise_for_status(status_code: int, content: str, response=None, request=None):
    if status_code == 403 and "Hydra token is either unexistent or revoked" in content:
        raise UserTokenExpired()
    elif status_code == 429:
        raise RateLimitError(f"Rate limited: {content}")
    elif status_code == 409:
        raise DuplicateError(f"Duplicate resource: {content}")
    elif status_code == 422:
        raise InvalidPayload(f"Invalid payload: {content}")
    elif status_code >= 400:
        if isinstance(response, HttpxResponse):
            raise HTTPStatusError(
                f"Status code {status_code}: {content}",
                request=request,
                response=response,
            )
        elif isinstance(response, RequestsResponse):
            raise RequestsHTTPError(
                f"Status code {status_code}: {content}", response=response
            )
        else:
            raise Exception(f"Status code {status_code}: {content}")


def serialize(obj):
    # Serialize each item in the iterator separately
    if hasattr(obj, "__iter__") and hasattr(obj, "__next__"):
        for i, item in enumerate(obj):
            serialized_item = serialize(item)
            if serialized_item:
                print(f"{item.__class__.__name__} {i}")
                print(serialized_item)
                print()
        return ""

    if isinstance(obj, ResourceList):
        data = []
        for resource in obj.resources:
            status = resource.metadata.status if resource.metadata is not None else ""

            cropped_title = (
                resource.title
                if not resource.title or len(resource.title) < MAX_TITLE_LEN
                else resource.title[: MAX_TITLE_LEN - 3] + "..."
            )
            data.append(
                [resource.id, resource.icon, cropped_title, status, resource.slug]
            )
        return tabulate(
            data,
            headers=["UUID", "Icon", "Title", "Status", "Slug"],
        )

    if isinstance(obj, KnowledgeBoxList):
        data = []
        for kb in obj.kbs:
            data.append([kb.uuid, kb.slug])
        return tabulate(
            data,
            headers=["UUID", "Slug"],
        )

    if isinstance(obj, TaskDefinition):
        return json.dumps(obj.validation, indent=2, ensure_ascii=False)

    if isinstance(obj, TaskList):
        output = ""
        if obj.tasks:
            output += "# Available tasks\n"
            output += "\n"
            data = []
            for task in obj.tasks:
                data.append([task.name, task.description])

            output += tabulate(
                data,
                headers=["Name", "Description"],
            )
            output += "\n\n"

        if obj.running:
            output += "# Running tasks\n"
            output += "\n"
            data = []
            for running_task in obj.running:
                data.append(
                    [
                        running_task.task.name,
                        running_task.id,
                        running_task.scheduled_at,
                    ]
                )

            output += tabulate(
                data,
                headers=["Task", "ID", "Started"],
            )
            output += "\n\n"

        if obj.configs:
            output += "# Configured tasks\n"
            output += "\n"
            data = []
            for running_task in obj.configs:
                data.append(
                    [
                        running_task.task.name,
                        running_task.id,
                    ]
                )

            output += tabulate(
                data,
                headers=["Task", "ID"],
            )
            output += "\n\n"
        return output

    if isinstance(obj, SyncAskResponse):
        if obj.status != "success":
            response = f"ERROR: {obj.error_details}"
        elif obj.answer_json is not None:
            response = (
                f"JSON: {json.dumps(obj.answer_json, indent=2, ensure_ascii=False)} \n"
            )
        else:
            response = f"Answer: {obj.answer} \n"

        if obj.metadata:
            response += f"Metadata: {obj.metadata.model_dump_json(indent=2)}\n"

        response += "Resources: \n"
        data = []
        for resource in obj.retrieval_results.resources.values():
            data.append([resource.id, resource.icon, resource.title])
        response += tabulate(
            data,
            headers=["UUID", "Icon", "Title"],
        )
        return response

    if isinstance(obj, ActivityLogsOutput):
        obj = obj.model_dump(exclude_unset=True)
    if isinstance(obj, AragAnswer):
        # Return formatted JSON string for CLI output
        arag_answer_dict = obj.model_dump(exclude_unset=True, exclude_none=True)
        # Show name instead of value so it is more understandable
        arag_answer_dict["operation"] = AnswerOperation(
            arag_answer_dict["operation"]
        ).name
        return json.dumps(arag_answer_dict, indent=2, ensure_ascii=False)
    return obj


def build_httpx_client(
    headers: dict[str, str] = {}, base_url: Optional[str] = None
) -> httpx.Client:
    return httpx.Client(
        headers={"User-Agent": USER_AGENT, **headers}, base_url=(base_url or "")
    )


def build_httpx_async_client(
    headers: dict[str, str] = {}, base_url: Optional[str] = None
) -> httpx.AsyncClient:
    return httpx.AsyncClient(
        headers={"User-Agent": USER_AGENT, **headers}, base_url=(base_url or "")
    )



import asyncio
import inspect
from functools import wraps

import yaml
from httpx import ConnectError

from nuclia import BASE_DOMAIN
from nuclia.data import (
    get_async_auth,
    get_async_client,
    get_auth,
    get_client,
)
from nuclia.exceptions import NotDefinedDefault, NucliaConnectionError
from nuclia.lib.agent import AgentClient, AsyncAgentClient
from nuclia.lib.kb import AsyncNucliaDBClient, Environment, NucliaDBClient
from nuclia.lib.nua import AsyncNuaClient, NuaClient


def accounts(func):
    @wraps(func)
    def wrapper_checkout_accounts(*args, **kwargs):
        auth = get_auth()
        auth.accounts()
        return func(*args, **kwargs)

    async def async_wrapper_checkout_accounts(*args, **kwargs):
        auth = get_async_auth()
        await auth.accounts(cached=False)
        result = await func(*args, **kwargs)
        return result

    if asyncio.iscoroutinefunction(func):
        return async_wrapper_checkout_accounts
    else:
        return wrapper_checkout_accounts


def kb(func):
    @wraps(func)
    async def async_wrapper_checkout(*args, **kwargs):
        if "ndb" in kwargs:
            return await func(*args, **kwargs)
        url = kwargs.get("url")
        api_key = kwargs.get("api_key")
        auth = get_async_auth()
        if url is None:
            # Get default KB
            kbid = auth._config.get_default_kb()
            if kbid is None:
                raise NotDefinedDefault()
            ndb = await get_async_client(kbid)
        elif url.find(BASE_DOMAIN) >= 0:
            region = url.split(".")[0].split("/")[-1]
            ndb = AsyncNucliaDBClient(
                environment=Environment.CLOUD, url=url, api_key=api_key, region=region
            )
        else:
            ndb = AsyncNucliaDBClient(environment=Environment.OSS, url=url)
        kwargs["ndb"] = ndb
        try:
            result = await func(*args, **kwargs)
            return result
        except ConnectError:
            raise NucliaConnectionError(f"Could not connect to {ndb}")

    @wraps(func)
    def wrapper_checkout(*args, **kwargs):
        if "ndb" in kwargs:
            return func(*args, **kwargs)
        url = kwargs.get("url")
        api_key = kwargs.get("api_key")
        auth = get_auth()
        if url is None:
            # Get default KB
            kbid = auth._config.get_default_kb()
            if kbid is None:
                raise NotDefinedDefault()
            ndb = get_client(kbid)
        elif url.find(BASE_DOMAIN) >= 0:
            region = url.split(".")[0].split("/")[-1]
            ndb = NucliaDBClient(
                environment=Environment.CLOUD, url=url, api_key=api_key, region=region
            )
        else:
            ndb = NucliaDBClient(environment=Environment.OSS, url=url)
        kwargs["ndb"] = ndb
        try:
            result = func(*args, **kwargs)
            return result
        except ConnectError:
            raise NucliaConnectionError(f"Could not connect to {ndb}")

    @wraps(func)
    async def async_generative_wrapper_checkout(*args, **kwargs):
        if "ndb" in kwargs:
            async for value in func(*args, **kwargs):
                yield value
        else:
            url = kwargs.get("url")
            api_key = kwargs.get("api_key")
            auth = get_async_auth()
            if url is None:
                # Get default KB
                kbid = auth._config.get_default_kb()
                if kbid is None:
                    raise NotDefinedDefault()
                ndb = await get_async_client(kbid)
            elif url.find(BASE_DOMAIN) >= 0:
                region = url.split(".")[0].split("/")[-1]
                ndb = AsyncNucliaDBClient(
                    environment=Environment.CLOUD,
                    url=url,
                    api_key=api_key,
                    region=region,
                )
            else:
                ndb = AsyncNucliaDBClient(environment=Environment.OSS, url=url)
            kwargs["ndb"] = ndb
            async for value in func(*args, **kwargs):
                yield value

    if inspect.isasyncgenfunction(func):
        return async_generative_wrapper_checkout
    elif asyncio.iscoroutinefunction(func):
        return async_wrapper_checkout
    else:
        return wrapper_checkout


def nucliadb(func):
    @wraps(func)
    def wrapper_checkout_nucliadb(*args, **kwargs):
        if "ndb" in kwargs:
            return func(*args, **kwargs)
        url = kwargs.get("url")
        auth = get_auth()
        if url is None:
            # Get default KB
            nucliadb = auth._config.get_default_nucliadb()
            if nucliadb is None:
                raise NotDefinedDefault()
            # OSS
            ndb = NucliaDBClient(
                environment=Environment.OSS, base_url=f"{nucliadb}/api"
            )
        else:
            ndb = NucliaDBClient(environment=Environment.OSS, base_url=f"{url}/api")
        kwargs["ndb"] = ndb
        return func(*args, **kwargs)

    return wrapper_checkout_nucliadb


def nua(func):
    @wraps(func)
    async def async_wrapper_checkout_nua(*args, **kwargs):
        if "nc" in kwargs:
            return await func(*args, **kwargs)
        auth = get_auth()
        nua_id = auth._config.get_default_nua()
        if nua_id is None:
            raise NotDefinedDefault()

        nua_obj = auth._config.get_nua(nua_id)
        nc = AsyncNuaClient(
            region=nua_obj.region, account=nua_obj.account, token=nua_obj.token
        )

        kwargs["nc"] = nc
        return await func(*args, **kwargs)

    @wraps(func)
    async def async_generative_wrapper_checkout_nua(*args, **kwargs):
        if "nc" in kwargs:
            async for value in func(*args, **kwargs):
                yield value
        else:
            auth = get_auth()
            nua_id = auth._config.get_default_nua()
            if nua_id is None:
                raise NotDefinedDefault()

            nua_obj = auth._config.get_nua(nua_id)
            nc = AsyncNuaClient(
                region=nua_obj.region, account=nua_obj.account, token=nua_obj.token
            )

            kwargs["nc"] = nc
            async for value in func(*args, **kwargs):
                yield value

    @wraps(func)
    def wrapper_checkout_nua(*args, **kwargs):
        if "nc" in kwargs:
            return func(*args, **kwargs)

        auth = get_auth()
        nua_id = auth._config.get_default_nua()
        if nua_id is None:
            raise NotDefinedDefault()

        nua_obj = auth._config.get_nua(nua_id)
        nc = NuaClient(
            region=nua_obj.region, account=nua_obj.account, token=nua_obj.token
        )

        kwargs["nc"] = nc
        return func(*args, **kwargs)

    if inspect.isasyncgenfunction(func):
        return async_generative_wrapper_checkout_nua
    elif asyncio.iscoroutinefunction(func):
        return async_wrapper_checkout_nua
    else:
        return wrapper_checkout_nua


def account(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        account_slug = kwargs.get("account")
        account_id = kwargs.get("account_id")
        auth = get_auth()
        if account_id is None and account_slug is None:
            account_slug = auth._config.get_default_account()
            if account_slug is None:
                raise NotDefinedDefault()
            kwargs["account"] = account_slug
        if account_id is None:
            account_id = auth.get_account_id(account_slug)  # type: ignore
            kwargs["account_id"] = account_id
        return func(*args, **kwargs)

    async def async_wrapper(*args, **kwargs):
        account_slug = kwargs.get("account")
        account_id = kwargs.get("account_id")
        auth = get_async_auth()
        if account_id is None and account_slug is None:
            account_slug = auth._config.get_default_account()
            if account_slug is None:
                raise NotDefinedDefault()
            kwargs["account"] = account_slug
        if account_id is None:
            account_id = auth.get_account_id(account_slug)  # type: ignore
            kwargs["account_id"] = account_id
        return await func(*args, **kwargs)

    if asyncio.iscoroutinefunction(func):
        return async_wrapper
    else:
        return wrapper


def pretty(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        if kwargs.get("json"):
            return result.json(indent=2)
        if kwargs.get("yaml"):
            return yaml.dump(result)
        return result

    @wraps(func)
    async def async_wrapper(*args, **kwargs):
        result = await func(*args, **kwargs)
        if kwargs.get("json"):
            return result.json(indent=2)
        if kwargs.get("yaml"):
            return yaml.dump(result)
        return result

    if asyncio.iscoroutinefunction(func):
        return async_wrapper
    else:
        return wrapper


def zone(func):
    @wraps(func)
    def wrapper_checkout_zone(*args, **kwargs):
        zone = kwargs.get("zone")
        if not zone:
            auth = get_auth()
            kwargs["zone"] = auth._config.get_default_zone()
        return func(*args, **kwargs)

    async def async_wrapper_checkout_zone(*args, **kwargs):
        zone = kwargs.get("zone")
        if not zone:
            auth = get_async_auth()
            kwargs["zone"] = auth._config.get_default_zone()
        return await func(*args, **kwargs)

    if asyncio.iscoroutinefunction(func):
        return async_wrapper_checkout_zone
    else:
        return wrapper_checkout_zone


def agent(func):
    def _get_agent_config(agent_id, auth):
        """Common logic to retrieve and validate agent configuration."""
        if agent_id is None:
            # Get default Agent
            agent_id = auth._config.get_default_agent()

        agent_obj = auth._config.get_agent(agent_id)
        if agent_obj is None:
            raise NotDefinedDefault()
        if agent_obj.region is None or agent_obj.account is None:
            raise ValueError(
                f"Agent {agent_id} is missing required fields (region or account)"
            )
        return agent_id, agent_obj

    def _create_agent_client(agent_id, agent_obj, auth, client_class):
        """Common logic to create an agent client."""
        return client_class(
            region=agent_obj.region,
            agent_id=agent_id,
            api_key=agent_obj.token,
            user_token=auth._config.token,
            account_id=agent_obj.account,
        )

    @wraps(func)
    def wrapper_checkout(*args, **kwargs):
        if "ac" in kwargs:
            return func(*args, **kwargs)

        auth = get_auth()
        agent_id, agent_obj = _get_agent_config(kwargs.get("agent_id"), auth)
        ac = _create_agent_client(agent_id, agent_obj, auth, AgentClient)
        kwargs["ac"] = ac

        try:
            result = func(*args, **kwargs)
            return result
        except ConnectError:
            raise NucliaConnectionError(f"Could not connect to {ac}")

    @wraps(func)
    async def async_wrapper_checkout(*args, **kwargs):
        if "ac" in kwargs:
            return await func(*args, **kwargs)

        auth = get_async_auth()
        agent_id, agent_obj = _get_agent_config(kwargs.get("agent_id"), auth)
        ac = _create_agent_client(agent_id, agent_obj, auth, AsyncAgentClient)
        kwargs["ac"] = ac

        try:
            result = await func(*args, **kwargs)
            return result
        except ConnectError:
            raise NucliaConnectionError(f"Could not connect to {ac}")

    @wraps(func)
    async def async_generative_wrapper_checkout(*args, **kwargs):
        if "ac" in kwargs:
            async for value in func(*args, **kwargs):
                yield value
        else:
            auth = get_async_auth()
            agent_id, agent_obj = _get_agent_config(kwargs.get("agent_id"), auth)
            ac = _create_agent_client(agent_id, agent_obj, auth, AsyncAgentClient)
            kwargs["ac"] = ac

            async for value in func(*args, **kwargs):
                yield value

    if inspect.isasyncgenfunction(func):
        return async_generative_wrapper_checkout
    elif asyncio.iscoroutinefunction(func):
        return async_wrapper_checkout
    else:
        return wrapper_checkout



import os
from datetime import datetime
from enum import Enum
from typing import List, Optional, Sequence
from urllib.parse import urlparse

from pydantic import BaseModel

from nuclia import CLOUD_ID
from nuclia.exceptions import NotDefinedDefault

CONFIG_DIR = "~/.nuclia"
CONFIG_PATH = CONFIG_DIR + "/config"


class KnowledgeBox(BaseModel):
    id: str
    url: str
    title: Optional[str] = None
    slug: Optional[str] = None
    token: Optional[str] = None
    region: Optional[str] = None
    account: Optional[str] = None

    def __str__(self):
        origin = self.region + " " + self.id if self.region else "LOCAL " + self.url
        return f"{origin:20} {'(' + self.account + ')' if self.account else ''} {self.slug if self.slug else ''}"


class User(BaseModel):
    name: str
    email: str
    type: str


class NuaKey(BaseModel):
    client_id: str
    account_type: Optional[str]
    region: str
    account: str
    token: str

    def __str__(self):
        return f"{self.client_id} {self.account} {self.account_type:30}"


class PersonalTokenCreate(BaseModel):
    id: str
    token: str
    expires: Optional[datetime]


class EphemeralToken(BaseModel):
    token: str


class PersonalTokenItem(BaseModel):
    id: str
    description: str
    expires: Optional[datetime]


class Zone(BaseModel):
    id: str
    title: str
    slug: Optional[str] = None

    def __str__(self):
        return f"{self.id:30} - {self.slug}"


class Account(BaseModel):
    id: str
    title: str
    slug: Optional[str] = None

    def __str__(self):
        return f"{self.id:30} - {self.slug:20} => {self.title:40}"


class Role(str, Enum):
    OWNER = "OWNER"
    READER = "READER"
    CONTRIBUTOR = "CONTRIBUTOR"


class Selection(BaseModel):
    nua: Optional[str] = None
    kbid: Optional[str] = None
    agent_id: Optional[str] = None
    account: Optional[str] = None
    nucliadb: Optional[str] = None
    zone: Optional[str] = None


def extract_region(url) -> Optional[str]:
    parsed = urlparse(url)
    hostname = parsed.hostname
    if hostname:
        parts = hostname.split(".")
        region = parts[0] if parts else None
        if region in [
            "localhost",
            "rag",
            "nuclia",
            "stashify",
            "gcp-global-dev-1",
            CLOUD_ID.split(".")[0],
        ]:
            # This means the URL is global, not regional
            return None
        return region
    return None


class Config(BaseModel):
    accounts: Optional[List[Account]] = []
    kbs: Optional[List[KnowledgeBox]] = []
    kbs_token: List[KnowledgeBox] = []
    agents: Optional[List[KnowledgeBox]] = []
    agents_token: List[KnowledgeBox] = []
    nuas_token: List[NuaKey] = []
    zones: Optional[List[Zone]] = []
    default: Optional[Selection] = Selection()
    user: Optional[str] = None
    token: Optional[str] = None

    def get_nua(self, nua_id: str) -> NuaKey:
        nua_obj = next(
            filter(
                lambda x: x.client_id == nua_id,
                self.nuas_token if self.nuas_token is not None else [],
            )
        )

        return nua_obj

    def get_kb(self, kbid: str) -> Optional[KnowledgeBox]:
        try:
            kb_obj = next(
                filter(
                    lambda x: x.id == kbid,
                    self.kbs_token if self.kbs_token is not None else [],
                )
            )
        except StopIteration:
            kb_obj = next(
                filter(lambda x: x.id == kbid, self.kbs if self.kbs is not None else [])
            )
        return kb_obj

    def get_agent(self, agent_id: str) -> Optional[KnowledgeBox]:
        try:
            agent_obj = next(
                filter(
                    lambda x: x.id == agent_id,
                    self.agents_token if self.agents_token is not None else [],
                )
            )
        except StopIteration:
            try:
                agent_obj = next(
                    filter(
                        lambda x: x.id == agent_id,
                        self.agents if self.agents is not None else [],
                    )
                )
            except StopIteration:
                agent_obj = None
        return agent_obj

    def set_user_token(self, code: str):
        self.token = code

    def remove_user_token(self):
        self.token = None
        self.save()

    def set_nua_token(
        self,
        client_id: str,
        account: str,
        base_region: str,
        token: str,
        account_type: Optional[str] = None,
    ):
        if self.nuas_token is None:
            self.nuas_token = []
        try:
            nua_obj = next(filter(lambda x: x.client_id == client_id, self.nuas_token))
            self.nuas_token.remove(nua_obj)
        except StopIteration:
            pass

        self.nuas_token.append(
            NuaKey(
                account_type=account_type,
                account=account,
                region=base_region,
                token=token,
                client_id=client_id,
            )
        )
        self.save()

    def _del_kbid(self, kbid: str):
        try:
            while True:
                kb_obj = next(
                    filter(
                        lambda x: x.id == kbid,
                        self.kbs_token if self.kbs_token is not None else [],
                    )
                )
                if self.kbs_token is not None:
                    self.kbs_token.remove(kb_obj)
        except StopIteration:
            pass

    def set_kb_token(
        self,
        url: str,
        kbid: str,
        token: Optional[str] = None,
        title: Optional[str] = None,
    ):
        self._del_kbid(kbid)
        region = extract_region(url)
        kb_obj = KnowledgeBox(id=kbid, url=url, token=token, title=title, region=region)
        self.kbs_token.append(kb_obj)

        if self.default is None:
            self.default = Selection()
            self.default.kbid = kbid

        self.save()

    def set_agent_token(
        self,
        url: str,
        agent_id: str,
        token: Optional[str] = None,
        title: Optional[str] = None,
        account_id: Optional[str] = None,
    ):
        # Remove existing agent with same ID
        try:
            while True:
                agent_obj = next(
                    filter(
                        lambda x: x.id == agent_id,
                        self.agents_token if self.agents_token is not None else [],
                    )
                )
                if self.agents_token is not None:
                    self.agents_token.remove(agent_obj)
        except StopIteration:
            pass

        region = extract_region(url)
        agent_obj = KnowledgeBox(
            id=agent_id,
            url=url,
            token=token,
            title=title,
            region=region,
            account=account_id,
        )
        self.agents_token.append(agent_obj)

        if self.default is None:
            self.default = Selection()
            self.default.agent_id = agent_id

        self.save()

    def get_default_nucliadb(self) -> Optional[str]:
        if self.default is None or self.default.nucliadb is None:
            raise NotDefinedDefault()
        return self.default.nucliadb

    def set_default_nucliadb(self, nucliadb: str):
        if self.default is None:
            self.default = Selection()
        self.default.nucliadb = nucliadb
        self.save()

    def get_default_nua(self) -> str:
        if self.default is None or self.default.nua is None:
            raise NotDefinedDefault()
        return self.default.nua

    def set_default_nua(self, nua: str):
        if self.default is None:
            self.default = Selection()
        self.default.nua = nua
        self.save()

    def get_default_account(self) -> str:
        if self.default is None or self.default.account is None:
            raise NotDefinedDefault()
        return self.default.account

    def set_default_account(self, account: str):
        if self.default is None:
            self.default = Selection()
        self.default.account = account
        self.save()

    def get_default_zone(self) -> Optional[str]:
        if self.default is None or self.default.zone is None:
            return None
        return self.default.zone

    def set_default_zone(self, zone: str):
        if self.default is None:
            self.default = Selection()
        self.default.zone = zone
        self.save()

    def get_default_kb(self) -> str:
        if self.default is None or self.default.kbid is None:
            raise NotDefinedDefault()
        return self.default.kbid

    def set_default_kb(self, kbid: str):
        if self.default is None:
            self.default = Selection()
        self.default.kbid = kbid
        self.save()

    def unset_default_kb(self, kbid: str):
        if self.default is None:
            self.default = Selection()
        if self.default.kbid == kbid:
            self.default.kbid = None
        self.save()

    def get_default_agent(self) -> str:
        if self.default is None or self.default.agent_id is None:
            raise NotDefinedDefault()
        return self.default.agent_id

    def set_default_agent(self, agent_id: str):
        if self.default is None:
            self.default = Selection()
        self.default.agent_id = agent_id
        self.save()

    def unset_default_agent(self, agent_id: str):
        if self.default is None:
            self.default = Selection()
        if self.default.agent_id == agent_id:
            self.default.agent_id = None
        self.save()

    def save(self):
        if not os.path.exists(os.path.expanduser(CONFIG_PATH)):
            os.makedirs(os.path.expanduser(CONFIG_DIR), exist_ok=True)

        from nuclia.data import DATA

        DATA.config = self
        with open(os.path.expanduser(CONFIG_PATH), "w") as config_file:
            config_file.write(self.model_dump_json())


def read_config() -> Config:
    if not os.path.exists(os.path.expanduser(CONFIG_PATH)):
        os.makedirs(os.path.expanduser(CONFIG_DIR), exist_ok=True)
        config = Config()
        with open(os.path.expanduser(CONFIG_PATH), "w") as config_file:
            config_file.write(config.model_dump_json())

    with open(os.path.expanduser(CONFIG_PATH), "r") as config_file:
        config = Config.model_validate_json(config_file.read())

    if config.default is None:
        config.default = Selection()
    return config


def retrieve(kbs: Sequence[KnowledgeBox], kb: str) -> Optional[KnowledgeBox]:
    kb_obj: Optional[KnowledgeBox] = None
    try:
        kb_obj = next(filter(lambda x: x.slug == kb, kbs))
    except StopIteration:
        try:
            kb_obj = next(filter(lambda x: x.id == kb, kbs))
        except StopIteration:
            pass
    return kb_obj


def retrieve_nua(nuas: List[NuaKey], nua: str) -> Optional[NuaKey]:
    nua_obj: Optional[NuaKey] = None
    try:
        nua_obj = next(filter(lambda x: x.client_id == nua, nuas))
    except StopIteration:
        pass
    return nua_obj


def retrieve_account(accounts: List[Account], account: str) -> Optional[Account]:
    account_obj: Optional[Account] = None
    try:
        account_obj = next(filter(lambda x: x.slug == account, accounts))
    except StopIteration:
        pass
    return account_obj


def set_config_file(path: str):
    global CONFIG_PATH
    CONFIG_PATH = path


def reset_config_file():
    global CONFIG_PATH
    CONFIG_PATH = CONFIG_DIR + "/config"
